/*
    Copyright (C) 2004 Kor Nielsen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <samplerate.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include FUSD_H_PATH

#include <sys/soundcard.h>
#include <jack/jack.h>
#include <sys/mman.h>
#include <sys/errno.h>


#define SUPPORTED_FORMATS AFMT_S16_LE | AFMT_S16_BE | AFMT_S16_NE | AFMT_S8 | AFMT_U8;

struct oss_instance
{
	int sample_format;
	int num_channels;
	int sample_rate;
	
	int samples_mixed;
	
	//current write info
	short* samples;
	short* sample_ptr;
	int samples_left_to_mix;
	int write_size;
	int buffer_size;
	
	int frag_size;
	int frag_count;
	
	float* buffer;
	float* wptr;
	float* rptr;
	
	double rate_ratio;
	
	struct SRC_STATE* rate_converter;
	
	SRC_DATA rate_data;
	
	struct fusd_file_info* file_info;
	
	struct oss_instance* next;
	
	struct fusd_file_info* polldiff;
};

struct device_info
{
	jack_port_t* output_ports[2];
	int sample_rate;
	pthread_mutex_t mutex;
	int samples_sent;
	struct oss_instance* instances;
	
};

void resize_buffer(struct oss_instance* instance, int buffer_size)
{
	if(buffer_size > instance->buffer_size)
	{
		if(instance->buffer)
			free(instance->buffer);
		instance->buffer = (float*)malloc((buffer_size + 1) * sizeof(float));
	}
	instance->buffer_size = buffer_size;
	instance->wptr = instance->buffer;
	instance->rptr = instance->buffer;
}

int get_sample_count(struct oss_instance* instance)
{
	int sample_count = instance->wptr - instance->rptr;
	if(sample_count < 0)
		sample_count += instance->buffer_size + 1;
	return sample_count;
}

int write_samples(struct oss_instance* instance, const short* samples, int num_samples)
{
	int max_samples = instance->buffer_size - get_sample_count(instance);
	if(num_samples > max_samples)
		num_samples = max_samples;
	
	if(instance->rate_converter == NULL)
	{
		const short* s = samples;
		const short* end = s + num_samples;
		float* d = instance->wptr;
		while(s < end)
		{
			*d++ = (*s++) * (1.0f / 32786.0f);
			if(d >= &instance->buffer[0] + instance->buffer_size + 1)
				d = instance->buffer;
		}
	}
	return num_samples;
}

int read_samples(struct oss_instance* instance, float* samples, int num_samples)
{
	float* s;
	float* d;
	float* end;
	
	int max_samples = get_sample_count(instance);
	if(num_samples > max_samples)
		num_samples = max_samples;
	
	s = instance->rptr;
	d = samples;
	end = samples + num_samples;
	
	while(d < end)
	{
		(*d++) = (*s++);
		if(s >= instance->buffer + instance->buffer_size + 1)
			s = instance->buffer;
	}
	return num_samples;
}

void skip_samples(struct oss_instance* instance, int num_samples)
{
	float* s;
	int total_samples = get_sample_count(instance);
	if(num_samples > total_samples)
		num_samples = total_samples;
		
	s = instance->rptr;
	s += num_samples;
	if(s >= instance->buffer + instance->buffer_size + 1)
		s -= instance->buffer_size + 1;
	
	instance->rptr = s;
}

void do_notify_complete(struct oss_instance* instance, struct device_info* dinfo);

int
jack_process (jack_nframes_t nframes, void *arg)
{
	struct device_info* info = (struct device_info*) arg;
	
	float* out_left = (float*) jack_port_get_buffer (info->output_ports[0], nframes);
	float* out_right = (float*) jack_port_get_buffer (info->output_ports[1], nframes);
	
	// clear jack buffers
	{
		float* out_left_end = out_left + nframes;
		float* out_l = out_left;
		float* out_r = out_right;
		
		while(out_l < out_left_end)
		{
			*out_l++ = 0;
			*out_r++ = 0;
		}
	}
	
	pthread_mutex_lock(&info->mutex);
	
	struct oss_instance* instance = info->instances;
	while(instance != NULL)
	{
		
		int count = get_sample_count(instance);
		
		float* s = instance->rptr;
		float* end = instance->buffer + instance->buffer_size + 1;
		float* out_left_end = out_left + nframes;
		float* out_l = out_left;
		float* out_r = out_right;
		
		if(instance->rate_converter != NULL)
		{
			int err;
			float in[1024];
			float out[1024];
			SRC_DATA data;

			data.end_of_input = 0;
			data.src_ratio = instance->rate_ratio;
			
			data.data_in = in;
			data.data_out = out;
			data.input_frames = read_samples(instance, in, 1024) / instance->num_channels;
			data.output_frames = nframes;
			
			if(data.output_frames > 512)
				data.output_frames = 512;
			
			err = src_process(instance->rate_converter, &data);
			
			if(err)
			{
				printf("error: %s\n", src_strerror(err));
				break;
			}
			
			s = data.data_out;
			
			out_left_end = out_l + data.output_frames_gen;
			skip_samples(instance, data.input_frames_used * instance->num_channels);

			if(instance->num_channels == 2)
			{
				while(out_l < out_left_end)
				{
					*out_l++ += *s++;
					*out_r++ += *s++;
				}
			}
			
			if(instance->num_channels == 1)
			{
				while(out_l < out_left_end)
				{
					*out_l++ += *s;
					*out_r++ += *s++;
				}
			}
			instance->samples_mixed +=  data.input_frames_used * instance->num_channels;
		}
		else
		{
			if(count > nframes)
				count = nframes;
				
			out_left_end = out_left + count;
			
			if(instance->num_channels == 2)
			{
				while(out_l < out_left_end)
				{
					*out_l++ += *s++;
						
					if(s >= end)
						s = instance->buffer;
							
					*out_r++ += *s++;
			
					if(s >= end)
						s = instance->buffer;
				}
			}
			
			if(instance->num_channels == 1)
			{
				while(out_l < out_left_end)
				{
					*out_l++ += *s;
						
					if(s >= end)
						s = instance->buffer;
							
					*out_r++ += *s++;
			
					if(s >= end)
						s = instance->buffer;
				}
			}
			instance->samples_mixed += count * instance->num_channels;
			instance->rptr = s;
		}
		instance = instance->next;
	}	
	
	info->samples_sent += nframes * 2;
	
	instance = info->instances;
	while(instance != NULL)
	{
		if(instance->sample_ptr != NULL)
		{
			int s_count = instance->samples_left_to_mix;
			short* s = instance->sample_ptr;
			float* d = instance->wptr;
			
			int ready_samples = s_count;
			int max_samples = instance->buffer_size - get_sample_count(instance);
			
			if(ready_samples > max_samples )
				ready_samples = max_samples;
			
			short* s_end = s + ready_samples;
			float* d_end = instance->buffer + instance->buffer_size + 1;
			
			
			while(s < s_end)
			{
				(*d++) = (1.0 / 32768.0f) * ((*s++));
	
				if(d >= d_end)
					d = instance->buffer;
			}

			
			instance->sample_ptr = s;
			instance->wptr = d;
			instance->samples_left_to_mix -= ready_samples;
			
			if(instance->samples_left_to_mix == 0)
			{
				instance->sample_ptr = NULL;
				fusd_return(instance->file_info, instance->write_size);
				instance->write_size = 0;
			}
		}
		do_notify_complete(instance, info);
		instance = instance->next;
	}
	
	pthread_mutex_unlock(&info->mutex);
	
	
	
	return 0;
}

int do_open(struct fusd_file_info* file)
{
	printf("Opened...\n");
	struct oss_instance* instance = (struct oss_instance*) malloc(sizeof(struct oss_instance));
	struct device_info* info = (struct device_info*) file->device_info;
	
	instance->sample_format = AFMT_S16_NE;
	instance->num_channels = 2;
	instance->sample_rate = info->sample_rate;
	instance->rate_converter = NULL;
	
	instance->samples = NULL;
	instance->sample_ptr = NULL;
	instance->samples_left_to_mix = 0;
	instance->samples_mixed = 0;
	instance->polldiff = NULL;
	instance->wptr = instance->buffer;
	instance->rptr = instance->buffer;
	instance->buffer_size = 0;
	instance->buffer = NULL;
	
	instance->frag_size = 4096;
	instance->frag_count = 8;
	
	resize_buffer(instance, 4096 * 8);
	
	file->private_data = instance;
	
	pthread_mutex_lock(&info->mutex);
	instance->next = info->instances;
	info->instances = instance;
	pthread_mutex_unlock(&info->mutex);
	
	return 0;
}

int do_close(struct fusd_file_info* file)
{
	printf("Closed...\n");
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct device_info* info = (struct device_info*) file->device_info;
	
	pthread_mutex_lock(&info->mutex);
	
	instance->wptr = instance->buffer;
	instance->rptr = instance->buffer;
	
	struct oss_instance* i = info->instances;
	while(i != NULL)
	{
		if(i->next == instance)
		{
			i->next = instance->next;
			break;
		}
		i = i->next;
	}
	if(info->instances == instance)
		info->instances = instance->next;
	pthread_mutex_unlock(&info->mutex);
	
	if(instance->rate_converter)
		src_delete(instance->rate_converter);
	
	if(instance->buffer)
		free(instance->buffer);
	
	if(instance != NULL)
		free(instance);
	
	return 0;
}

int do_read(struct fusd_file_info* file, char* buffer, size_t size, loff_t* offset)
{
	printf("Read %d bytes\n", size);
	return 0;
}

int do_write(struct fusd_file_info *file, const char *buffer,
             size_t size, loff_t *offset)
{
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct device_info* info = (struct device_info*) file->device_info;
	
	int sampleCount = (size >> 2) << 1;
	
	short* s = (short*) buffer;	
	
	
	
	pthread_mutex_lock(&info->mutex);
	
		int ready_samples = sampleCount;
		int max_samples = instance->buffer_size - get_sample_count(instance);
			
		if(ready_samples > max_samples )
			ready_samples = max_samples;
			
		short* s_end = s + ready_samples;
		float* d_end = instance->buffer + instance->buffer_size + 1;
		float* d = instance->wptr;
	
		while(s < s_end)
		{
			(*d++) = (1.0 / 32768.0f) * ((*s++));
		
			if(d >= d_end)
				d = instance->buffer;
		}

		instance->wptr = d;
	
	if(ready_samples < sampleCount)
	{
		instance->samples = (short*) buffer;
		instance->sample_ptr = s;
		instance->samples_left_to_mix = sampleCount - ready_samples;
		instance->write_size = sampleCount * 2;
		instance->file_info = file;
		pthread_mutex_unlock(&info->mutex);

		return -FUSD_NOREPLY;
	}
	pthread_mutex_unlock(&info->mutex);
	
	return ready_samples * 2;
}

void do_notify_complete(struct oss_instance* instance, struct device_info* dinfo)
{
	int curr_state, cached_state;
	
	if(instance == NULL || instance->polldiff == NULL)
	{
		return;
	}
	if(instance->buffer_size > get_sample_count(instance) )
		curr_state = FUSD_NOTIFY_OUTPUT;
	else
		curr_state = 0;
	
	cached_state = fusd_get_poll_diff_cached_state(instance->polldiff);
	
	if(curr_state != cached_state)
	{
		fusd_return(instance->polldiff, curr_state);
		instance->polldiff = NULL;
	}
}

ssize_t do_notify_polldiff(struct fusd_file_info* file, unsigned int cached_state)
{
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct device_info* info = (struct device_info*) file->device_info;
	
	if(instance == NULL)
		return -EINVAL;
	
	pthread_mutex_lock(&info->mutex);
	
	if(instance->polldiff != NULL)
	{
		fusd_destroy(instance->polldiff);
		instance->polldiff = NULL;
	}
	
	instance->polldiff = file;
	do_notify_complete(instance, info);
	
	pthread_mutex_unlock(&info->mutex);
	
	return -FUSD_NOREPLY;
}

int do_ioctl(struct fusd_file_info *file, int request, void *data)
{
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct device_info* dinfo = (struct device_info*) file->device_info;
	
	pthread_mutex_lock(&dinfo->mutex);
	int return_value = -1;
	switch(request)
	{
		case SNDCTL_DSP_RESET:
			instance->wptr = instance->buffer;
			instance->rptr = instance->buffer;
			//instance->samples_mixed = dinfo->samples_sent + 64 * 2;
			//instance->wptr = g_wptr + 64 * 2;
			//if(instance->wptr > g_wend)
			//	instance->wptr -= instance->buffer_size;
			//return 0;
			break;
		
		case SNDCTL_DSP_SETFMT:
			printf("SetFmt\n");
			return_value = 0;
			break;
		
		case SNDCTL_DSP_GETFMTS:
			printf("GetFmts\n");
			break;
		
		case SNDCTL_DSP_CHANNELS:
			{
				printf("SNDCTL_DSP_CHANNELS\n");
				int nChannels = *(int*)data;
				if(nChannels <= 0)
					*(int*)data = 1;
				else if(nChannels > 2)
					*(int*)data = 2;
				
				*(int*)data = instance->num_channels;
				printf("num_channels: %d\n", instance->num_channels);
				
				if(instance->rate_converter != NULL)
				{
					printf("Creating new rate_converter\n");
					int err;
					src_delete(instance->rate_converter);
					instance->rate_converter = src_new(SRC_SINC_FASTEST, instance->num_channels, &err);
				}
			}
			return_value = 0;
			break;
			
		case SNDCTL_DSP_STEREO:
			{
				printf("SNDCTL_DSP_STEREO %d\n", *(int*)data);
				int nChannels = *(int*)data + 1;
				if(nChannels <= 0)
					nChannels = 1;
				else if(nChannels > 2)
					nChannels = 2;
				
				instance->num_channels = nChannels;
				
				printf("num_channels: %d\n", instance->num_channels);
				
				if(instance->rate_converter != NULL)
				{
					int err;
					printf("Creating new rate_converter\n");
					src_delete(instance->rate_converter);
					instance->rate_converter = src_new(SRC_SINC_FASTEST, 1, &err);
				}
			}
			return_value = 0;
			break;
		
		case SNDCTL_DSP_SPEED:
			{
				int err = 0;
				
				if(instance->rate_converter == NULL && (*(int*)data) != dinfo->sample_rate)
				{
					printf("Rate converter channels: %d\n", instance->num_channels);
					instance->rate_converter = src_new(SRC_LINEAR, instance->num_channels, &err);
					if(err)
					{
						*(int*)data = instance->sample_rate;
						return_value = 0;
					}
				}
				instance->rate_ratio = (double)dinfo->sample_rate / (double)(*(int*)data);
				instance->sample_rate = *(int*)data;
				printf("Rate ratio: %eq\n", instance->rate_ratio);
				//src_set_ratio(instance->rate_converter, instance->rate_ratio);
			}
			return_value = 0;
			break;
			
		case SNDCTL_DSP_SYNC:
			printf("Sync\n");
			break;
		
		
		case SNDCTL_DSP_GETBLKSIZE:
			printf("SNDCTL_DSP_GETBLKSIZE\n");
			*((int*)data) = (instance->frag_size) * 2;
			return_value = 0;
			break;
		
		case SOUND_PCM_WRITE_FILTER:
			printf("SOUND_PCM_WRITE_FILTER\n");
			break;
		
		case SNDCTL_DSP_POST:
			printf("SNDCTL_DSP_POST\n");
			break;
			
		case SNDCTL_DSP_SUBDIVIDE:
			printf("SNDCTL_DSP_SUBDIVIDE\n");
			break;
		
		case SNDCTL_DSP_SETFRAGMENT:
			printf("SNDCTL_DSP_SETFRAGMENT %d\n", (*(int*)data) & 0xffff);
			instance->frag_size = (1 << ((*(int*)data) & 0xffff)) / 2;
			instance->frag_count = ((*(int*)data) & 0xffff0000) >> 16;
			
			if(instance->frag_count > 8)
				instance->frag_count = 8;
			
			(*(int*)data) = ((*(int*)data) & 0xffff ) | instance->frag_count << 16;
			
			printf("FragSize: %d x %d\n", instance->frag_size, instance->frag_count);
			resize_buffer(instance, instance->frag_size * instance->frag_count);
			return_value = 0;
			break;
		
		case SNDCTL_DSP_GETOSPACE:
		{
			struct audio_buf_info* info = (struct audio_buf_info*) data;
			info->fragments = (instance->buffer_size - get_sample_count(instance)) / (instance->buffer_size/instance->frag_count);
			if(info->fragments > instance->frag_count)
				info->fragments = instance->frag_count;
			info->fragstotal = instance->frag_count;
			info->fragsize = (instance->buffer_size / instance->frag_count) * 2;
			info->bytes = (instance->buffer_size - get_sample_count(instance)) * 2;
			//printf("DSP_GETOSPACE: fragstotal: %d  fragsize: %d   bytes: %d\n", info->fragstotal, info->fragsize, info->bytes);
			return_value = 0;
			break;
		}

		case SNDCTL_DSP_GETISPACE:
			printf("SNDCTL_DSP_GETISPACE\n");
			break;
		
		case SNDCTL_DSP_NONBLOCK:
			printf("SNDCTL_DSP_NONBLOCK\n");
			break;
		
		case SNDCTL_DSP_GETCAPS:
		printf("SNDCTL_DSP_GETCAPS\n");
			*(int*)data = 0;
			return_value = 0;
			break;
		
		case SNDCTL_DSP_GETTRIGGER:
			printf("SNDCTL_DSP_GETTRIGGER\n");
			break;
		
		case SNDCTL_DSP_SETTRIGGER:
			printf("SNDCTL_DSP_SETTRIGGER\n");
			return_value = 0;
			break;
			
			case SNDCTL_DSP_GETIPTR:
			printf("SNDCTL_DSP_GETIPTR\n");
			break;
			
		case SNDCTL_DSP_GETOPTR:
			((count_info*)data)->bytes = instance->samples_mixed * 2;
			
			((count_info*)data)->blocks = 0;
			((count_info*)data)->ptr = 0;
			
			return_value = 0;
			break;

		case SNDCTL_DSP_MAPINBUF:
			printf("SNDCTL_DSP_MAPINBUF\n");
			break;
			
		case SNDCTL_DSP_MAPOUTBUF:
			printf("SNDCTL_DSP_MAPOUTBUF\n");
			break;
			
		case SNDCTL_DSP_SETSYNCRO:
			printf("SNDCTL_DSP_SETSYNCRO\n");
			break;
			
		case SNDCTL_DSP_GETODELAY:
			//printf("SNDCTL_DSP_GETODELAY\n");
			*((int*)data) = get_sample_count(instance) * 2 + 128 * 4;
			return_value = 0;
			break;
			
		case SNDCTL_DSP_GETCHANNELMASK:
			printf("SNDCTL_DSP_GETCHANNELMASK\n");
			break;
			
		case SNDCTL_DSP_BIND_CHANNEL:
			printf("SNDCTL_DSP_BIND_CHANNEL\n");
			break;
		
		default:
			printf("ioctl %d\n", request);
			break;
	}
	
	pthread_mutex_unlock(&dinfo->mutex);
	
	return return_value;
}

struct fusd_file_operations oss_file_ops = 
{
	open: do_open,
	read: do_read,
	write: do_write,
	close: do_close,
	ioctl: do_ioctl,
	poll_diff: do_notify_polldiff
};

int
jack_srate (jack_nframes_t nframes, void *arg)

{
	struct device_info* info = (struct device_info*) arg;
	info->sample_rate = nframes;
	return 0;
}

void
jack_error (const char *desc)
{
	fprintf (stderr, "JACK error: %s\n", desc);
}

void
jack_shutdown (void *arg)
{
	exit (1);
}

void setscheduler(void)
{
        struct sched_param sched_param;

        if (sched_getparam(0, &sched_param) < 0) {
                printf("Scheduler getparam failed...\n");
                return;
        }
        sched_param.sched_priority = 10;
        if (!sched_setscheduler(0, SCHED_FIFO, &sched_param)) {
                printf("Scheduler set to Round Robin with priority %i...\n", sched_param.sched_priority);
                fflush(stdout);
                return;
        }
        printf("!!!Scheduler set to Round Robin with priority %i FAILED!!!\n", sched_param.sched_priority);
}


int main(int argc, char *argv[])
{
	mlockall(MCL_CURRENT | MCL_FUTURE);
	setscheduler();
	
	jack_client_t* client;
	struct device_info* info;
	const char** ports;
	
	jack_set_error_function(jack_error);
	
	client = jack_client_new("oss-dsp");
	
	if(client == NULL)
	{
		printf("Error registering jack client. Make sure that jackd is running.\n");
		return -1;
	}
	
	info = (struct device_info*) malloc(sizeof(struct device_info));
	
	pthread_mutex_init(&info->mutex, NULL);
	info->instances = NULL;
	info->samples_sent = 0;
	
	info->output_ports[0] = jack_port_register(client, "left", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
	info->output_ports[1] = jack_port_register(client, "right", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
	
	jack_set_process_callback(client, jack_process, info);
	jack_set_sample_rate_callback(client, jack_srate, info);
	jack_on_shutdown(client, jack_shutdown, info);
	
	jack_activate(client);

	ports = jack_get_ports (client, NULL, NULL, JackPortIsPhysical|JackPortIsInput);
	
	jack_connect (client, jack_port_name(info->output_ports[0]), ports[0]);
	jack_connect (client, jack_port_name(info->output_ports[1]), ports[1]);

	if(fusd_register("/dev/sound/dsp", 0666, info, &oss_file_ops) < 0)
	{
		perror("Could not create sound device\n");
		return 1;
	}
	
	printf("Running...\n");
	fusd_run();
	
	pthread_mutex_destroy(&info->mutex);
	return 0;
}
